#include "hash.h"

extern const char LOG_STRING[17][32];

/*
 * the local functions for doing queue manipulations
 * (adding buffers and executing them accordingly)
 */
static void execute(char *);
static int try_execute(char *);
static int get_waitfor(char *buffer);
static int get_command(char *buffer);
static int get_depend_num(char *);
static void append_q_ele(q_t *, char *);
static q_t *create_q(int, char *);
static void push_q(q_t *, char *);
static void add_buffer_q(q_list_t *, int, char *);

/*
 * the local functions for doing stack manipulations
 * (roll back of buffers and undoing them properly)
 */
static void undo(char *buffer);
static void roll_back_q(q_t *q);
static void roll_back_list(q_list_t *l);

/*
 * create a hash table and return it
 */
hash_t *init_hash(void) {
  int i;
  hash_t *h = (hash_t *)malloc(sizeof(hash_t));

  for(i = 0; i < QHASHSIZE; i++) {
    h->list[i].head = NULL;
    h->list[i].tail = NULL;
  }

  return h;
}

/*
 * destroy a hash table
 */
void free_hash(hash_t *h) {
  free(h);
}


/*
 * execute the buffer command
 */
void execute(char *buffer) {
  parse_buffer(buffer, BUF_EXECUTE);
  free(buffer);
}

/*
 * see what the waitfor value of the buffer is
 * if it has none, then simply execute the buffer now
 */
int try_execute(char *buffer) {
  int waitfor = get_waitfor(buffer);
  
  if(!waitfor) {
    /* if we don't have to wait, don't! */
    execute(buffer);
  }

  return waitfor;
}

/*
 * get the type of command for while this buffer should wait
 */
int get_waitfor(char *buffer) {
  return parse_buffer(buffer, BUF_GET_WAITFOR);
}

/*
 * return the command number related to a given buffer
 */
int get_command(char *buffer) {
  return ((int *)(buffer))[0];
}

/*
 * return the inode number related to a given buffer
 */
int get_depend_num(char *buffer) {
  return parse_buffer(buffer, BUF_GET_DEPENDENT_VAL);
}

/*
 * create a queue element and put it at the end of the queue
 */
void append_q_ele(q_t *q, char *buffer) {
  q_ele_t *ele = (q_ele_t *)malloc(sizeof(q_ele_t));
  if(ele == NULL) {
    return;
  }
  ele->buffer = buffer;
  ele->next = NULL;
  ele->prev = q->tail;
  if(q->tail) {
    q->tail->next = ele;
  } else {
    q->head = ele;
  }
  q->tail = ele;
}

/*
 * either create a queue with a single element in it
 * or execute the command and return NULL
 */
q_t *create_q(int dep, char *buffer) {
  int waitfor;
  q_t *q = NULL;
  q_ele_t *ele;

  waitfor = try_execute(buffer);
  if(waitfor) {
    /* didn't get to execute, so we have to wait */
    q = (q_t *)malloc(sizeof(q_t));
    if(q == NULL) {
      /* error, couldn't allocate memory! */
      /* note: no good error handling here */
      return NULL;
    }
    q->inode = dep;
    q->tail = NULL;
    q->head = NULL;
    q->wait_command = waitfor;
    append_q_ele(q, buffer);
  }
  return q;
}

/*
 * push the buffer onto the queue, if it's the command
 * we've been waiting for, execute the entire queue
 */
void push_q(q_t *q, char *buffer) {
  q_ele_t *ele;
  int command = get_command(buffer);

  /*
   * note: there is a special case where one of two commands are possible
   * exits, however, there may be better ways to deal with this, and if
   * other cases such as this arise, I would suggest trying to find one.
   */
  if(command == q->wait_command ||
     (command == LOGGING_CHANGESIZE &&
      q->wait_command == LOGGING_FREEINODE)) {
    /* execute the queue */
    while(q->head) {
      ele = q->head;
      execute(ele->buffer);
      q->head = ele->next;
      free(ele);
    }
    /* and the passed in buffer */
    execute(buffer);
    q->tail = NULL;
  } else {
    /* drop the buffer on the end of the queue */
    append_q_ele(q, buffer);
  }
}

/*
 * attempt add a buffer to a queue on the queue list,
 * or execute that buffer
 */
void add_buffer_q(q_list_t *l, int dep, char *buffer) {
  /* find the proper queue */
  q_t *ptr = l->head;
  while(ptr && ptr->inode != dep) {
    ptr = ptr->next;
  }

  if(!ptr) {
    /* must be a new inode, start a queue */
    ptr = create_q(dep, buffer);

    if(!ptr) {
      /* just got executed, nothing needed */
    } else {
      /* put this item at the end of the list */
      ptr->prev = l->tail;
      ptr->next = NULL;
      if(l->tail != NULL) {
	l->tail->next = ptr;
      } else {
	l->head = ptr;
      }
      l->tail = ptr;
    }
  } else {
    /* found the proper queue, push this item on */
    push_q(ptr, buffer);
    
    if(ptr->head == NULL) {
      /* looks like the queue was executed, we can now remove it */
      if(!ptr->prev && !ptr->next) {
	l->head = NULL;
	l->tail = NULL;
      } else if(!ptr->next) {
	l->tail = ptr->prev;
	ptr->prev->next = ptr->next;
      } else if(!ptr->prev) {
	l->head = ptr->next;
	ptr->next->prev = ptr->prev;
      } else {
	ptr->prev->next = ptr->next;
	ptr->next->prev = ptr->prev;
      }
      free(ptr);
    }
  }
}

/*
 * add a buffer to the proper hashed queue
 */
void add_buffer(hash_t *h, char *buffer) {
  int dep;

  /* get the inode number from the buffer */
  dep = get_depend_num(buffer);

  /* add the buffer to the appropriate list */
  add_buffer_q(&(h->list[HASHDEP(dep)]), dep, buffer);
}




/*
 * undo a single buffer operation
 */
void undo(char *buffer) {
  parse_buffer(buffer, BUF_UNDO);
}

/*
 * roll back a single queue
 */
void roll_back_q(q_t *q) {
  q_ele_t *ele;

  while(q->tail) {
    ele = q->tail;
    undo(ele->buffer);
    q->tail = ele->prev;
    free(ele);
  }
  q->head = NULL;
}

/*
 * roll back all the queues in this list
 */
void roll_back_list(q_list_t *l) {
  q_t *ptr;

  while(l->head) {
    ptr = l->head;
    roll_back_q(ptr);
    l->head = ptr->next;
    free(ptr);
  }
  l->tail = NULL;
}

/*
 * go through all hashed queues and roll back any changes left incompleted 
 */
void roll_back(hash_t *h) {
  int i;

  for(i = 0; i < QHASHSIZE; i++) {
    roll_back_list(&(h->list[i]));
  }
}

/*
 * initialize a list of blocks
 */
list_t *init_list() {
  list_t *l = (list_t *)malloc(sizeof(list_t));
  if(l) {
    l->head = NULL;
    l->tail = NULL;
  }
  return l;
}

/*
 * put data from a block onto a list
 */
void put_block_list(list_t *l, int bno) {
  list_ele_t *ele;
  struct bufarea *bp;
  struct fs *fs = &sblock;

  /* retrieve the data */
  ele = (list_ele_t *)malloc(sizeof(list_ele_t));
  if(!ele) {
    /* need better error handling! */
    return;
  }
  bp = getdatablk(bno, fs->fs_bsize);
  ele->bno = bno;
  ele->data = (char *)malloc(fs->fs_bsize);
  if(!ele->data) {
    /* need better error handling! */
    free(ele->data);
    return;
  }
  memcpy(ele->data, bp->b_un.b_buf, fs->fs_bsize);
  
  /* zero out the buffer */
  bzero(bp->b_un.b_buf, fs->fs_bsize);
  dirty(bp);
  bp->b_flags &= ~B_INUSE;

  /* put it on the end of the list */
  ele->prev = l->tail;
  ele->next = NULL;
  if(l->tail) {
    l->tail->next = ele;
  } else {
    l->head = ele;
  }
  l->tail = ele;
}

/*
 * get data from a list and put it into a block
 */
void get_block_list(list_t *l, int bno) {
  list_ele_t *ele;
  struct bufarea *bp;
  struct fs *fs = &sblock;

  /* retrieve the data from the list */
  ele = l->head;
  while(ele && ele->bno != bno) {
    ele = ele->next;
  } if(!ele) {
    /* better error handling? (maybe not here, since this is often the case) */
    return;
  }
  
  /* remove the element from the list */
  if(!ele->next && !ele->prev) {
    l->tail = ele->prev;
    l->head = ele->next;
  } else if(!ele->next) {
    l->tail = ele->prev;
    ele->prev->next = ele->next;
  } else if(!ele->prev) {
    l->head = ele->next;
    ele->next->prev = ele->prev;
  } else {
    ele->prev->next = ele->next;
    ele->next->prev = ele->prev;
  }

  /* put the data into the block */
  bp = getdatablk(bno, fs->fs_bsize);
  memcpy(bp->b_un.b_buf, ele->data, fs->fs_bsize);
  dirty(bp);
  bp->b_flags &= ~B_INUSE;

  /* free the element */
  free(ele->data);
  free(ele);
}

/*
 * free a list of blocks
 */
void free_list(list_t *l) {
  list_ele_t *ele = l->head;

  while(l->head) {
    l->head = ele->next;
    free(ele->data);
    free(ele);
    ele = l->head;
  }
  free(l);
}
