/* Craig Soules
   soules@andrew.cmu.edu
   Apr, 1999 */

#include <ufs/ffs/ffs_logging.h>
#include <sys/resourcevar.h>
#include <sys/resource.h>
#include <sys/proc.h>

/* local prototypes */
static void logging_checkpoint_log(struct mount *, int,
				   struct fs *, struct vnode *);
static void logging_log_buffer(struct mount *, char *, int, struct inode *);
static void update_buf(struct mount *, struct vnode *, int);
static void update_this_buf(struct mount *, struct buf *);
int log_checksum(int *, int);

/*
 * given a buffer, find out how much log to write out
 * and write it out
 */
void logging_flush_log_to_buf(bp)
     struct buf *bp;
{
  unsigned int offset;
  int start_block, end_block, block_count;
  struct buf *log_bp;
  struct inode *ino_p;
  struct vnode *log_file_p;
  struct mount *mp = bp->b_log_mount;

  /* get the far end of the log to be written out */
  offset = bp->b_log_write;

  /* check if this buffer is the super-block of the filesystem */
  if(bp->b_log_fsflag) {
    /* if so, then we want to update the fs->log_start to co-inside with the
     * mp->log_checkstart */
    struct fs *fs = (struct fs *)(bp->b_data);

    fs->fs_log_start = mp->mnt_log_checkstart;
    ((struct ufsmount *)(mp->mnt_data))->um_fs->fs_log_start = mp->mnt_log_checkstart;

    /*
     * recompute the size of the log
     */
    if(mp->mnt_log_checkstart > mp->mnt_log_end) {
      mp->mnt_log_disksize =
	mp->mnt_log_maxsize - mp->mnt_log_checkstart + mp->mnt_log_end;
    } else {
      mp->mnt_log_disksize = mp->mnt_log_end - mp->mnt_log_checkstart;
    }

    bp->b_log_mount = NULL;
    bp->b_log_fsflag = 0;
  } /* note: mp can only be set if the fs is a logging fs */
  else if(mp != NULL) {
    lockmgr(&mp->mnt_loglock_flush, LK_EXCLUSIVE, NULL, NULL);
    /* if we are allowed to lock the log 
       if(!bp->b_log_checkpointing) {
	  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);
	  }
    */

    /* reset the buffer for future use */
    bp->b_log_write = 0;
    bp->b_log_check = 0;
    bp->b_log_mount = NULL;

    /* take the buffer out of the list of related buffers */
    if(bp->b_log_next == NULL && bp->b_log_prev == NULL) {
      mp->mnt_log_head = NULL;
      mp->mnt_log_tail = NULL;
      /* no blocks depending on the log in memory! the log isn't needed! */
      mp->mnt_log_checkstart = mp->mnt_log_end;
      mp->mnt_log_size = 0;
    } else if(bp->b_log_next == NULL) {
      bp->b_log_prev->b_log_next = NULL;
      mp->mnt_log_tail = bp->b_log_prev;
    } else if(bp->b_log_prev == NULL) {
      bp->b_log_next->b_log_prev = NULL;
      mp->mnt_log_head = bp->b_log_next;
      
      /* we are checkpointing some stuff, move the mp->log_checkstart foward */
      mp->mnt_log_checkstart = bp->b_log_next->b_log_check;
      if(mp->mnt_log_end >= mp->mnt_log_checkstart) {
	mp->mnt_log_size = mp->mnt_log_end - mp->mnt_log_checkstart;
      } else {
	mp->mnt_log_size = mp->mnt_log_maxsize - mp->mnt_log_checkstart +
	  mp->mnt_log_end;
      }
    } else {
      bp->b_log_next->b_log_prev = bp->b_log_prev;
      bp->b_log_prev->b_log_next = bp->b_log_next;
    }
    bp->b_log_next = NULL;
    bp->b_log_prev = NULL;

    if((mp->mnt_log_end > mp->mnt_log_current && 
	(offset > mp->mnt_log_current && offset <= mp->mnt_log_end)) ||
       (mp->mnt_log_end < mp->mnt_log_current &&
	(offset > mp->mnt_log_current || offset <= mp->mnt_log_end))) {
      /* flush any buffers floating in memory that have data needed in the
       * disk log to the disk */

      /* open the file */
      if(log_file_p = ufs_ihashlookup(VFSTOUFS(mp)->um_dev, LOGGINGINO)) {
	ino_p = VTOI(log_file_p);
	start_block = mp->mnt_log_current / ino_p->i_fs->fs_bsize;
	end_block = offset / ino_p->i_fs->fs_bsize;
	if(start_block == end_block && mp->mnt_log_current > offset) {
	  start_block++;
	}
	if(end_block >= start_block) {
	  block_count = end_block - start_block + 1;
	} else {
	  block_count = (mp->mnt_log_maxsize / ino_p->i_fs->fs_bsize) -
	    start_block + end_block + 1;
	}

	while(block_count > 0) {
	  if(log_bp = gbincore(log_file_p, start_block)) {
	    if(log_bp->b_flags & (B_DIRTY|B_DELWRI)) {
	      /*************************/
	      if(!(log_bp->b_flags & B_BUSY)) {
		if(log_bp->b_qindex != QUEUE_NONE) {
		  bremfree(log_bp);
		}
		log_bp->b_flags |= B_BUSY;
		log_bp->b_flags &= ~B_ASYNC;
		/*		log_bp->b_flags |= B_CLUSTEROK; */
		bwrite(log_bp);
		/*
		 * note: I'd like to do some sort of clustered write
		 *       here, it should increase efficiency
		 */
	      } else {
		biowait(log_bp);
	      }
	    }
	  }
	  start_block = (start_block+1) %
	    (mp->mnt_log_maxsize / ino_p->i_fs->fs_bsize);
	  block_count--;
	}

	/* update the current end of the log on disk */
	mp->mnt_log_current = offset;
      }
    }

    lockmgr(&mp->mnt_loglock_flush, LK_RELEASE, NULL, NULL);
    /* if we were allowed to lock the log 
       if(!bp->b_log_checkpointing) {
	  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
	  } else {
	  bp->b_log_checkpointing = 0;
	  }
    */
  }

  return;
}

/*
 * trim the log file
 */
static void logging_checkpoint_log(mp, size, fs, vp)
     struct mount *mp;
     int size;
     struct fs *fs;
     struct vnode *vp;
{
  struct buf *bp;
  int s, flags;
  
  s = splbio();

  /* force the checkpointing of some buffers */
  while((mp->mnt_log_size + size) > mp->mnt_log_maxsize) {
    /* get the buffer pointer to the first thing to checkpoint */
    bp = mp->mnt_log_head;

    /*
     * I'm not certain why this happens... however, if the buffer is invalid,
     * we can't exactly write the data out, so I just flush all the related
     * things using the same call that would be made if the data was written
     * and then continue on as normal
     */       
    if(bp->b_flags & B_INVAL) {
      logging_flush_log_to_buf(bp);
      continue;
    }

    /* write it out... the rest will be handled by the i/o logging hook */
    /*************************/
    if(!(bp->b_flags & B_BUSY)) {
      if(bp->b_qindex != QUEUE_NONE) {
	bremfree(bp); 
      }
      bp->b_flags |= B_BUSY;
      flags = bp->b_flags & B_ASYNC;
      bp->b_flags &= ~(B_ASYNC);
      bp->b_log_checkpointing = 1;
      splx(s);
      bwrite(bp);
      bp->b_log_checkpointing = 0;
      s = splbio();
      bp->b_flags |= flags;
    } else {
      /* since the buffer is in use, we don't want to confuse
	 anything, so do a forced write of the buffer without
	 moving it around or messing with it */
      bp->b_flags &= ~(B_READ | B_DONE | B_ERROR | B_DELWRI);
      flags = bp->b_flags;

      /* do this? the block has to be dirty... */
      --numdirtybuffers;
      reassignbuf(bp, bp->b_vp);

      bp->b_flags |= B_WRITEINPROG | B_BUSY;
      bp->b_vp->v_numoutput++;
      vfs_busy_pages(bp, 1);

      bp->b_log_checkpointing = 2;
      splx(s);
      VOP_STRATEGY(bp->b_vp, bp);
      mp->mnt_stat.f_syncwrites++;
      biowait(bp);
      bp->b_log_checkpointing = 0;
      s = splbio();
      bp->b_flags = flags;
    }
  }


  /* write the super block */
  if(bp = gbincore(vp, SBLOCK)) {
    /* set the superblock flags */
    bp->b_log_fsflag = 1;
    bp->b_log_mount = mp;

    if(!(bp->b_flags & B_BUSY)) {
      /* if the buffer isn't busy... bwrite() */
      if(bp->b_qindex != QUEUE_NONE) {
	bremfree(bp);
      }
      bp->b_flags |= B_BUSY;
      flags = bp->b_flags & B_ASYNC;
      bp->b_flags &= ~(B_ASYNC);
      bp->b_log_checkpointing = 3;
      splx(s);
      bwrite(bp);
      bp->b_flags |= flags;
      bp->b_log_checkpointing = 0;
    } else {
      /* otherwise, still need to flush the buffer, but can't bwrite() */
      bp->b_flags &= ~(B_READ | B_DONE | B_ERROR | B_DELWRI);
      flags = bp->b_flags;

      /* do this? the block has to be dirty... */
      --numdirtybuffers;
      reassignbuf(bp, bp->b_vp);
      
      bp->b_flags |= B_WRITEINPROG | B_BUSY;
      bp->b_vp->v_numoutput++;
      vfs_busy_pages(bp, 1);
      
      bp->b_log_checkpointing = 4;
      splx(s);
      VOP_STRATEGY(bp->b_vp, bp);
      mp->mnt_stat.f_syncwrites++;
      biowait(bp);
      bp->b_log_checkpointing = 0;
      bp->b_flags = flags;
    }
  } else {
    /* grab the superblock */
    splx(s);
    bread(vp, SBLOCK, fs->fs_bsize, NOCRED, &bp);
    s = splbio();

    /* set the superblock flag */
    bp->b_log_fsflag = 1;
    bp->b_log_mount = mp;

    /* write it out! (updating will occur during the write) */
    if(bp->b_qindex != QUEUE_NONE) {
      bremfree(bp);
    }
    bp->b_flags |= B_BUSY;
    splx(s);
    bwrite(bp);
  }
}

static void logging_log_buffer(mp, buffer, size, ip)
     struct mount *mp;
     char *buffer;
     int size;
     struct inode *ip;
{
  struct vnode *log_file_p;
  struct buf *log_bp;
  int offa, offb;
  int blocka, blockb;
  int part1, part2, part3;

  /* checkpoint if necessary */
  if((mp->mnt_log_disksize + size) > mp->mnt_log_maxsize) {
    logging_checkpoint_log(mp, size, ip->i_fs, ip->i_devvp);
  }

  /* grab the vnode */
  ffs_vget(mp, LOGGINGINO, &log_file_p);

  /* find the appropriate buffer(s) and do the copy */
  offa = mp->mnt_log_end;
  offb = (mp->mnt_log_end + size) % mp->mnt_log_maxsize;
  blocka = offa / VTOI(log_file_p)->i_fs->fs_bsize;
  blockb = offb / VTOI(log_file_p)->i_fs->fs_bsize;

  if(blocka == blockb) {
    part1 = mp->mnt_log_end % VTOI(log_file_p)->i_fs->fs_bsize;
    bread(log_file_p, blocka, VTOI(log_file_p)->i_fs->fs_bsize,
	  NOCRED, &log_bp);
    memcpy(log_bp->b_data+part1, buffer, size);
    bdwrite(log_bp);
  } else if(blockb > blocka) {
    part1 = (blockb * VTOI(log_file_p)->i_fs->fs_bsize) - mp->mnt_log_end;
    part2 = size - part1;
    part3 = VTOI(log_file_p)->i_fs->fs_bsize - part1;

    bread(log_file_p, blocka, VTOI(log_file_p)->i_fs->fs_bsize,
	  NOCRED, &log_bp);
    memcpy(log_bp->b_data + part3, buffer, part1);
    bdwrite(log_bp);

    bread(log_file_p, blockb, VTOI(log_file_p)->i_fs->fs_bsize,
	  NOCRED, &log_bp);
    memcpy(log_bp->b_data, buffer+part1, part2);
    bdwrite(log_bp);
  } else {
    part1 = mp->mnt_log_maxsize - mp->mnt_log_end;
    part2 = size - part1;
    part3 = VTOI(log_file_p)->i_fs->fs_bsize - part1;

    bread(log_file_p, blocka, VTOI(log_file_p)->i_fs->fs_bsize,
	  NOCRED, &log_bp);
    memcpy(log_bp->b_data + part3, buffer, part1);
    bdwrite(log_bp);

    bread(log_file_p, blockb, VTOI(log_file_p)->i_fs->fs_bsize,
	  NOCRED, &log_bp);
    memcpy(log_bp->b_data, buffer+part1, part2);
    bdwrite(log_bp);
  }

  /* update the different pointers */
  mp->mnt_log_size += size;
  mp->mnt_log_disksize += size;
  mp->mnt_log_end = (mp->mnt_log_end + size) % mp->mnt_log_maxsize;

  /* drop the vnode */
  vput(log_file_p);
}

int logging_flushlog(oldmnt, flags, p)
     struct mount *oldmnt;
     int flags;
     struct proc *p;
{
  struct vnode *log_file_p;

#ifdef DEBUG_LOGGING
  printf("Flush the log ->\n");
  printf("\tLoop, writing all the log information to the disk\n");
  printf("\tDo we need to do a file sync here?\n");
  printf("\tRelease the vnode on /.log\n");
#endif

  /* lock the log file */
  lockmgr(&oldmnt->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* open the file */
  ffs_vget(oldmnt, LOGGINGINO, &log_file_p);

  /* syncing the log? is this necessary? */
  
  /* release the vnode */
  vput(log_file_p);

  /* unlock the log file */
  lockmgr(&oldmnt->mnt_loglock, LK_RELEASE, NULL, NULL);

  return 0;
}

int logging_mount(devvp, mp, fs, cred)
	struct vnode *devvp;
	struct mount *mp;
	struct fs *fs;
	struct ucred *cred;
{
  struct vnode *vp;

  ffs_vget(mp, LOGGINGINO, &vp);

#ifdef DEBUG_LOGGING
  printf("Mount the file system as a logging ffs ->\n");
  printf("\tDo any extra special stuff to prepare for logging here!\n");
#endif

  lockinit(&mp->mnt_loglock, PVFS, "vfslock", 0, LK_NOPAUSE);
  lockinit(&mp->mnt_loglock_flush, PVFS, "vfslock", 0, LK_NOPAUSE);
  mp->mnt_flag |= MNT_LOGGING;
  mp->mnt_log_maxsize = VTOI(vp)->i_size;
  mp->mnt_log_checkstart = 0;
  mp->mnt_log_end = 0;
  mp->mnt_log_size = 0;
  mp->mnt_log_disksize = 0;
  mp->mnt_log_head = NULL;
  mp->mnt_log_tail = NULL;
  VTOI(vp)->i_fs->fs_log_start = 0;
  VTOI(vp)->i_fs->fs_log_end = 0;

  vput(vp);

  return 0;
}

void logging_initialize() {
  printf("Initialize all logging stuff here!\n");
}

static void update_buf(mp, vp, blkno)
     struct mount *mp;
     struct vnode *vp;
     int blkno;
{
  struct buf *bp;

  if(bp = gbincore(vp, blkno)) {
    /* update the last place in the log this points to */
    bp->b_log_write = mp->mnt_log_end;
    if(!bp->b_log_mount) {
      /* update the first place in the log this points to */
      bp->b_log_check = mp->mnt_log_end;

      /* put the buffer on the end of the list, since it has the newest 
       * log entry */
      bp->b_log_next = NULL;
      bp->b_log_prev = mp->mnt_log_tail;
      if(bp->b_log_prev != NULL) {
	bp->b_log_prev->b_log_next = bp;
      } else {
	mp->mnt_log_head = bp;
      }
      mp->mnt_log_tail = bp;

      /* set the mount pointer */
      bp->b_log_mount = mp;
    }
  }
}

static void update_this_buf(mp, bp)
     struct mount *mp;
     struct buf *bp;
{
  if(bp) {
    /* update the last place in the log this points to */
    bp->b_log_write = mp->mnt_log_end;
    if(!bp->b_log_mount) {
      /* update the first place in the log this points to */
      bp->b_log_check = mp->mnt_log_end;

      /* put the buffer on the end of the list, since it has the newest 
       * log entry */
      bp->b_log_next = NULL;
      bp->b_log_prev = mp->mnt_log_tail;
      if(bp->b_log_prev != NULL) {
	bp->b_log_prev->b_log_next = bp;
      } else {
	mp->mnt_log_head = bp;
      }
      mp->mnt_log_tail = bp;

      /* set the mount pointer */
      bp->b_log_mount = mp;
    }
  }
}

/*
 * create a checksum of the passed in values
 */
int log_checksum(int *buf, int size) {
  int cksum = 0, i;

  for(i = 0; i < size; i++) {
    cksum += buf[i];
  }
  return cksum;
}

/*
 *  call this when you want to create a direct pointer block
 */
void logging_setup_direct(ip, lbn, newblkno, newsize, bp)
     struct inode *ip;	        /* inode to which block is being added */
     ufs_lbn_t lbn;		/* block pointer within inode */
     ufs_daddr_t newblkno;	/* disk block number being added */
     long newsize;		/* size of new block */
     struct buf *bp;            /* dep if the block is an indir ptr block */
{
  int buffer[7];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Creating a direct pointer ->\n");
  printf("\tinode number: %d\n", ip->i_number);
  printf("\tblock pointer within inode: %d\n", (int)lbn);
  printf("\tblock number being added: %d\n", (int)newblkno);
  printf("\tthe new size of the file?: %d\n", (int)newsize);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode number
   * - logical block number to allocate
   * - actual block number added
   * - checksum
   */
  buffer[0] = LOGGING_ADDDIRECT;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = lbn;
  buffer[5] = newblkno;
  buffer[6] = log_checksum(buffer, 6);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*7, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));

  /*
   * if the inode is a directory, then we need to make a dependency
   * for the new directory block ? */
  update_this_buf(mp, bp);

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

/* 
 *  call this when you want to update the used block map
 */
void logging_blkmapdep(ip, mnt, newblkno, frags)
     struct inode *ip;          /* the inode that is getting blocks */
     struct mount *mnt;         /* the mount point structure */
     ufs_daddr_t newblkno;	/* number of newly allocated block */
     int frags;                 /* the number of frags being added */
{
  int buffer[7];
  struct timespec time;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Allocating a block ->\n");
  printf("\tadding frags at block number: %d\n", newblkno);
  printf("\tadding %d frags\n", frags);
#endif

  /* lock the log */
  lockmgr(&mnt->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - the inode number that is getting these blocks
   * - block number
   * - number of frags to allocate
   * - checksum
   */
  buffer[0] = LOGGING_UPDATEBLKMAP;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = newblkno;
  buffer[5] = frags;
  buffer[6] = log_checksum(buffer, 6);

  /* make the log changes */
  logging_log_buffer(mnt, (char *)buffer, sizeof(int)*7, ip);

  /* update the associated buffers */
  update_buf(mnt, ip->i_devvp, fsbtodb(fs, cgtod(fs, dtog(fs, newblkno))));

  /* update the inode buf? */

  /* unlock the log */
  lockmgr(&mnt->mnt_loglock, LK_RELEASE, NULL, NULL);
}

/*
 *  call this when you want to update the used inode map
 */
void logging_inomapdep(ip, newinum, mode)
     struct inode *ip;		/* inode related to allocation */
     ino_t newinum;		/* new inode number being allocated */
     int mode;			/* the mode of the inode */
{
  int buffer[6];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Allocating an inode ->\n");
  printf("\tupdate the inode map with this new inode: %d\n", newinum);
  printf("\talso note the type of the inode: %d\n", ip->i_mode);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode number
   * - inode mode bits
   * - checksum
   */
  buffer[0] = LOGGING_UPDATEINOMAP;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = newinum;
  buffer[4] = mode;
  buffer[5] = log_checksum(buffer, 5);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*6, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, cgtod(fs, ino_to_cg(fs, newinum))));

  update_buf(mp, ip->i_devvp, fsbtodb(fs, ino_to_fsba(fs, newinum)));

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

/*
 * call this when you want to free a file
 */
void logging_freefile(pvp, ino, mode)
     struct vnode *pvp;
     ino_t ino;
     int mode;
{
  int buffer[5];
  struct timespec time;
  struct mount *mp = pvp->v_mount;
  struct fs *fs = VTOI(pvp)->i_fs;

#ifdef DEBUG_LOGGING
  printf("Freeing an inode ->\n");
  printf("\tzeroed inode %d written to disk with all 0's\n", ino);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode number
   * - checksum
   */
  buffer[0] = LOGGING_FREEINODE;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ino;
  buffer[4] = log_checksum(buffer, 4);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*5, VTOI(pvp));

  /* update the associated buffers */
  update_buf(mp, VTOI(pvp)->i_devvp,
	     fsbtodb(fs, cgtod(fs, ino_to_cg(fs, ino))));

  update_buf(mp, VTOI(pvp)->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ino)));

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}


/*
  call this when you want to update information in an inode

  THIS IS WHERE THE INODE IS TOSSED FROM THE CACHE!!!!!

  not so certain about this now... might just be the update for the
  modified and updated times... but is definitely written to disk here...
*/
void logging_update_inodeblock(ip, bp, waitfor)
     struct inode *ip;	/* the "in_core" copy of the inode */
     struct buf *bp;		/* the buffer containing the inode block */
     int waitfor;		/* 1 => update must be allowed */
{
  /*
    printf("Updating the inode ->\n");
    printf("\tneed to note the access/modified/inode change times?\n");
  */
}

void logging_init_directory(ip, dp, bp)
     struct inode *ip;
     struct inode *dp;
     struct buf *bp;
{
  char buffer[sizeof(int)*8];
  struct timespec time;
  struct mount *mp = ITOV(dp)->v_mount;
  struct fs *fs = dp->i_fs;

#ifdef DEBUG_LOGGING
  printf("Init a directory block ->\n");
  printf("\tdirectory's inode: %d\n", ip->i_number);
  printf("\tdir link: %d; parent link: %d\n", ip->i_nlink, dp->i_nlink);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - directory inode
   * - directory link count
   * - parent inode
   * - parent link count
   * - checksum
   * - directory entries
   */
  ((int *)(buffer))[0] = LOGGING_INITDIR;
  ((int *)(buffer))[1] = time.tv_sec;
  ((int *)(buffer))[2] = time.tv_nsec;
  ((int *)(buffer))[3] = ip->i_number;
  ((int *)(buffer))[4] = ip->i_nlink;
  ((int *)(buffer))[5] = dp->i_number;
  ((int *)(buffer))[6] = dp->i_nlink;
  ((int *)(buffer))[7] = log_checksum((int *)buffer, 7);

  /* make the log changes */
  logging_log_buffer(mp, buffer, sizeof(int)*8, dp);

  /* update the associated buffers */
  update_buf(mp, dp->i_devvp, fsbtodb(fs, ino_to_fsba(fs, dp->i_number)));

  update_buf(mp, dp->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));

  /***** dirp *****/
  update_this_buf(mp, bp);

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

void logging_setup_directory_add(dp, prev_reclen, dirp, diroffset, dbp, i_link)
     struct inode *dp;		/* inode for directory */
     unsigned int prev_reclen;	/* notes how large the previous entry is */
     struct direct *dirp;	/* inode referenced by new directory entry */
     off_t diroffset;		/* offset of new entry in directory */
     struct buf *dbp;           /* buffer that holds the directory entry */
     int i_link;                /* the new link count of the inode */
{
  char buffer[sizeof(int)*9 + 8 + 256];
  struct timespec time;
  struct mount *mp = ITOV(dp)->v_mount;
  struct fs *fs = dp->i_fs;

#ifdef DEBUG_LOGGING
  printf("Add a directory entry for a given inode ->\n");
  printf("\tdirectory's inode: %d\n", dp->i_number);
  printf("\tat offset: %d\n", (int)diroffset);
  printf("\tprevious entry's new reclen: %d\n", prev_reclen);
  printf("\tdlink: %d; ilink: %d\n", dp->i_nlink, i_link);
  if(dirp != NULL) {
    printf("\tinode number: %d\n", dirp->d_ino);
    printf("\tnamlen: %d\n", dirp->d_namlen);
  } else {
    printf("\tinode number: %d\n", -1);
  }
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - parent directory inode
   * - parent directory link count
   * - inode link count
   * - offset point in directory
   * - previous record length
   * - directory entry
   * - checksum
   */
  ((int *)(buffer))[0] = LOGGING_ADDDIRENTRY;
  ((int *)(buffer))[1] = time.tv_sec;
  ((int *)(buffer))[2] = time.tv_nsec;
  ((int *)(buffer))[3] = dp->i_number;
  ((int *)(buffer))[4] = dp->i_nlink;
  ((int *)(buffer))[5] = i_link;
  ((int *)(buffer))[6] = diroffset;
  ((int *)(buffer))[7] = prev_reclen;
  ((int *)(buffer))[8] = log_checksum((int *)buffer, 8);
  memcpy(buffer + sizeof(int)*9, dirp, 8 + dirp->d_namlen);

  /* make the log changes */
  logging_log_buffer(mp, buffer, sizeof(int)*9 + 8 + dirp->d_namlen, dp);

  /* update the associated buffers */
  update_buf(mp, dp->i_devvp, fsbtodb(fs, ino_to_fsba(fs, dp->i_number)));

  update_buf(mp, dp->i_devvp, fsbtodb(fs, ino_to_fsba(fs, dirp->d_ino)));

  /***** dirp *****/
  update_this_buf(mp, dbp);

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}


void logging_setup_remove(dp, ip, bp)
     struct inode *dp;		/* inode for the directory being modified */
     struct inode *ip;		/* inode for directory entry being removed */
     struct buf *bp;
{
  int buffer[11];
  struct timespec time;
  struct mount *mp = ITOV(dp)->v_mount;
  struct fs *fs = dp->i_fs;

#ifdef DEBUG_LOGGING
  printf("Removing a directory entry ->\n");
  printf("\tdirectory inode: %d\n", dp->i_number);
  printf("\tentry inode: %d\n", ip->i_number);
  printf("\tdlink: %d; ilink: %d\n", dp->i_nlink, ip->i_nlink);
  printf("\toffset into block: %d\n", dp->i_offset);
  printf("\tcount of previous entry: %d\n", dp->i_count);
  printf("\trecord length of entry being removed: %d\n", dp->i_reclen);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - parent directory inode
   * - parent directory link count
   * - the inode number
   * - inode link count
   * - offset of inode being removed
   * - offset of entry before inode being removed
   * - size of inode being removed
   * - checksum
   */
  buffer[0] = LOGGING_REMDIRENTRY;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = dp->i_number;
  buffer[4] = dp->i_nlink;
  buffer[5] = ip->i_number;
  buffer[6] = ip->i_nlink;
  buffer[7] = dp->i_offset;
  buffer[8] = dp->i_count;
  buffer[9] = dp->i_reclen;
  buffer[10] = log_checksum(buffer, 10);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*11, dp);

  /* update the associated buffers */
  /* note: is this inode truely a dependency? i don't think so... */
  update_buf(mp, dp->i_devvp, fsbtodb(fs, ino_to_fsba(fs, dp->i_number)));

  /* the directory entry block here! */
  update_this_buf(mp, bp);

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

void logging_setup_directory_change(dp, ip, newinum, ilink, bp)
     struct inode *dp;	 /* inode for the directory being modified */
     struct inode *ip;	 /* inode for directory entry being removed */
     long newinum;	 /* new inode number for changed entry */
     int ilink;          /* link count of the new inode number */
     struct buf *bp;     /* buffer associated with the directory entry block */
{
  int buffer[11];
  struct timespec time;
  struct mount *mp = ITOV(dp)->v_mount;
  struct fs *fs = dp->i_fs;

#ifdef DEBUG_LOGGING
  printf("Changing a directory entry ->\n");
  printf("\tdir inode: %d; new inode: %d", dp->i_number, (int)newinum);
  printf("\tdir nlink: %d; new nlink: %d\n", dp->i_nlink, ilink);
  printf("\tentry inode offset: %d\n", dp->i_offset);
  printf("\told inode: %d\n", (int)ip->i_number);
  printf("\told inode link count: %d\n", ip->i_nlink);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - directory inode
   * - directory inode link count
   * - old inode number
   * - old inode linke count
   * - offset of inode being changed
   * - new inode to change it to
   * - new inode link count
   * - checksum
   */
  buffer[0] = LOGGING_CHANGEDIRENTRY;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = dp->i_number;
  buffer[4] = dp->i_nlink;
  buffer[5] = ip->i_number;
  buffer[6] = ip->i_nlink;
  buffer[7] = dp->i_offset;
  buffer[8] = newinum;
  buffer[9] = ilink;
  buffer[10] = log_checksum(buffer, 10);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*11, dp);

  /* update the associated buffers */
  update_buf(mp, dp->i_devvp, fsbtodb(fs, ino_to_fsba(fs, dp->i_number)));

  update_buf(mp, dp->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));

  /***** dirp *****/
  update_this_buf(mp, bp);

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

void logging_move_directory(dp, prev_reclen, dirp, diroffset, dbp, i_link)
     struct inode *dp;		/* inode for directory */
     unsigned int prev_reclen;	/* notes how large the previous entry is */
     struct direct *dirp;	/* inode referenced by new directory entry */
     off_t diroffset;		/* offset of new entry in directory */
     struct buf *dbp;           /* buffer that holds the directory entry */
     int i_link;                /* the new link count of the inode */
{
  char buffer[sizeof(int)*9 + 8 + 256];
  struct timespec time;
  struct mount *mp = ITOV(dp)->v_mount;
  struct fs *fs = dp->i_fs;

#ifdef DEBUG_LOGGING
  printf("Moving a directory entry for a given inode ->\n");
  printf("\tdirectory's inode: %d\n", dp->i_number);
  printf("\tat offset: %d\n", (int)diroffset);
  printf("\tprevious entry's new reclen: %d\n", prev_reclen);
  printf("\tdlink: %d; ilink: %d\n", dp->i_nlink, i_link);
  if(dirp != NULL) {
    printf("\tinode number: %d\n", dirp->d_ino);
    printf("\tnamlen: %d\n", dirp->d_namlen);
  } else {
    printf("\tinode number: %d\n", -1);
  }
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - parent directory inode
   * - parent directory link count
   * - inode link count
   * - offset point in directory
   * - previous record length
   * - directory entry
   * - checksum
   */
  ((int *)(buffer))[0] = LOGGING_MOVDIRENTRY;
  ((int *)(buffer))[1] = time.tv_sec;
  ((int *)(buffer))[2] = time.tv_nsec;
  ((int *)(buffer))[3] = dp->i_number;
  ((int *)(buffer))[4] = dp->i_nlink;
  ((int *)(buffer))[5] = i_link;
  ((int *)(buffer))[6] = diroffset;
  ((int *)(buffer))[7] = prev_reclen;
  ((int *)(buffer))[8] = log_checksum((int *)buffer, 8);
  memcpy(buffer + sizeof(int)*9, dirp, 8 + dirp->d_namlen);

  /* make the log changes */
  logging_log_buffer(mp, buffer, sizeof(int)*9 + 8 + dirp->d_namlen, dp);

  /* update the associated buffers */
  update_buf(mp, dp->i_devvp, fsbtodb(fs, ino_to_fsba(fs, dp->i_number)));

  update_buf(mp, dp->i_devvp, fsbtodb(fs, ino_to_fsba(fs, dirp->d_ino)));

  /***** dirp *****/
  update_this_buf(mp, dbp);

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

/*
 * Called just before setting an indirect block pointer to a
 * newly allocated indirect block.
 */
void logging_setup_indir_meta(ip, ptrno, parentblkno, newblkno)
     struct inode *ip;	/* inode for file being extended */
     int ptrno;		/* offset of pointer in indirect block */
     ufs_daddr_t parentblkno;   /* block number of the parent pointer block */
     ufs_daddr_t newblkno;	/* block number being added */
{
  int buffer[8];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Creating a new indirect pointer ->\n");
  printf("inode number: %d\n", ip->i_number);
  printf("pointer of indirect block: %d\n", ptrno);
  printf("disk block number being added: %d\n", newblkno);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode number
   * - offset into indirect block
   * - block number of parent indirect block
   * - block number of new indirect block
   * - checksum
   */
  buffer[0] = LOGGING_SETINDIRMETA;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = ptrno;
  buffer[5] = parentblkno;
  buffer[6] = newblkno;
  buffer[7] = log_checksum(buffer, 7);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*8, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, ptrno));

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

/*
 * Called just before setting an indirect block pointer
 * to a newly allocated file page.
 */
void logging_setup_indir_page(ip, lbn, newblkno, newsize)
     struct inode *ip;	/* inode for file being extended */
     ufs_lbn_t lbn;		/* allocated block number within file */
     ufs_daddr_t newblkno;	/* disk block number being added */
     int newsize;
{
  int buffer[8];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Creating a new indirect page ->\n");
  printf("\tallocated block number: %d\n", (int)lbn);
  printf("\tblock number being added: %d\n", (int)newblkno);
  printf("\tnew size: %d\n", newsize);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode being extended
   * - logical block number being allocated
   * - the block number allocated to it
   * - previous block number to free if any
   * - checksum
   */
  buffer[0] = LOGGING_ADDINDIRPAGE;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = lbn;
  buffer[5] = newblkno;
  buffer[6] = newsize;
  buffer[7] = log_checksum(buffer, 7);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*8, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

void logging_change_size(ip, size)
     struct inode *ip;
     int size;
{
  int buffer[7];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Changing the size of a file ->\n");
  printf("\tinode number: %d\n", ip->i_number);
  printf("\tinode size: %d\n", size);
  printf("\tnumber of blocks %d\n", ip->i_blocks);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode number
   * - new size
   * - correct number of blocks
   * - checksum
   */
  buffer[0] = LOGGING_CHANGESIZE;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = size;
  buffer[5] = ip->i_blocks;
  buffer[6] = log_checksum(buffer, 6);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*7, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

void logging_setup_freefrags(ip, bno, size, bp)
     struct inode *ip;
     ufs_daddr_t bno;
     long size;
     struct buf *bp;
{
  int buffer[7];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Freeing frags ->\n");
  printf("\tblock number: %d\n", bno);
  printf("\tnumber of frags: %d\n", (int)(size / ip->i_fs->fs_fsize));
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - the inode these frags are associated with
   * - block number
   * - number of fragments to free
   * - checksum
   */
  buffer[0] = LOGGING_FREEFRAGS;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = bno;
  buffer[5] = (size / fs->fs_fsize);
  buffer[6] = log_checksum(buffer, 6);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*7, ip);

  /* update the associated buffers */
  /* ip->i_devvp, fsbtodb(fs, cgtod(fs, dtog(fs, bno))) */
  update_this_buf(mp, bp);

  /* update the inode buf? */

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

void logging_setup_freeblocks(ip, length)
     struct inode *ip;	/* The inode whose length is to be reduced */
     off_t length;	/* The new length for the file */
{
  int buffer[7];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Freeing blocks ->\n");
  printf("\tinode being deleted: %d\n", ip->i_number);
  printf("\told size: %d\n", (int)ip->i_size);
  printf("\tnew size: %d\n", (int)length);
  printf("\talso keep a list of all the associated blocks\n");
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode being deleted
   * - old size
   * - new size
   * - checksum
   */
  buffer[0] = LOGGING_FREEBLOCKS;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = ip->i_size;
  buffer[5] = length;
  buffer[6] = log_checksum(buffer, 6);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*7, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));

  /* need to make dependencies for any indirect blocks */
  

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

void logging_symlink(ip, link, len)
     struct inode *ip;
     char *link;
     int len;
{
  char buffer[sizeof(int)*6 + (NDADDR + NIADDR)*sizeof(int)];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Logging the symlink creation ->\n");
  printf("\tinode of the symlink: %d\n", ip->i_number);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode number
   * - length of symlink
   * - symlink name
   */
  ((int *)(buffer))[0] = LOGGING_SYMLINK;
  ((int *)(buffer))[1] = time.tv_sec;
  ((int *)(buffer))[2] = time.tv_nsec;
  ((int *)(buffer))[3] = ip->i_number;
  ((int *)(buffer))[4] = len;
  ((int *)(buffer))[5] = log_checksum((int *)buffer, 5);
  memcpy(buffer + sizeof(int)*6, link, len);

  /* make the log changes */
  logging_log_buffer(mp, buffer, sizeof(int)*6 + len, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

void logging_setattr(ip)
     struct inode *ip;
{
  int buffer[9];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Logging the attribute change ->\n");
  printf("\tinode of the change: %d\n", ip->i_number);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode number
   * - inode mode bits
   * - inode uid
   * - inode gid
   * - inode flags
   * - checksum
   */
  buffer[0] = LOGGING_SETATTR;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = ip->i_mode;
  buffer[5] = ip->i_uid;
  buffer[6] = ip->i_gid;
  buffer[7] = ip->i_flags;
  buffer[8] = log_checksum((int *)buffer, 8);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*9, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

void logging_mknod(ip)
     struct inode *ip;
{
  int buffer[6];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Logging the device node creation ->\n");
  printf("\tinode of the device node: %d\n", ip->i_number);
  printf("\trdev of the device node: %d\n", ip->i_rdev);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode number
   * - inode device
   * - checksum
   */
  buffer[0] = LOGGING_MKNOD;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = ip->i_rdev;
  buffer[5] = log_checksum((int *)buffer, 5);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*6, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

void logging_realloc_buf(ip, bp, bno, osize, nsize)
     struct inode *ip;
     struct buf *bp;
     ufs_daddr_t bno;
     int osize;
     int nsize;
{
  int buffer[9];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;

#ifdef DEBUG_LOGGING
  printf("Logging the reallocation of a block ->\n");
  printf("\tinode: %d\n", ip->i_number);
  printf("\toriginal block: %d\n", dbtofsb(fs, bp->b_blkno));
  printf("\tnew block: %d\n", bno);
  printf("\tsize to copy: %d\n", osize);
  printf("\ttotal size (zero the rest): %d\n", nsize);
#endif

  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);

  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode number
   * - inode device
   * - checksum
   */
  buffer[0] = LOGGING_REALLOCBUF;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = dbtofsb(fs, bp->b_blkno);
  buffer[5] = bno;
  buffer[6] = osize;
  buffer[7] = nsize;
  buffer[8] = log_checksum((int *)buffer, 8);

  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*9, ip);
  logging_log_buffer(mp, bp->b_data, osize, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));
  update_this_buf(mp, bp);

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

/************************** END FILE HERE! *********************/
/* THIS FUNCTION IS PROBABLY OBSOLETE, BUT I'M KEEPING TEMP    */
/* STORAGE OF IT HERE FOR NOW...                               */
/***************************************************************/


void logging_increase_linkcnt(ip)
     struct inode *ip;	/* the inode with the increased link count */
{
  int buffer[6];
  struct timespec time;
  struct mount *mp = ITOV(ip)->v_mount;
  struct fs *fs = ip->i_fs;
  
#ifdef DEBUG_LOGGING
  printf("Increasing the link count ->\n");
  printf("\tinode nunber: %d; %d\n", ip->i_number, ip->i_nlink);
#endif
  
  /* lock the log */
  lockmgr(&mp->mnt_loglock, LK_EXCLUSIVE, NULL, NULL);
  
  /* get the time */
  getnanotime(&time);

  /* layout:
   *
   * - Entry number
   * - timestamp (2)
   * - inode number
   * - link count
   * - checksum
   */
  buffer[0] = LOGGING_INCLINKCNT;
  buffer[1] = time.tv_sec;
  buffer[2] = time.tv_nsec;
  buffer[3] = ip->i_number;
  buffer[4] = ip->i_nlink;
  buffer[5] = log_checksum(buffer, 5);
  
  /* make the log changes */
  logging_log_buffer(mp, (char *)buffer, sizeof(int)*6, ip);

  /* update the associated buffers */
  update_buf(mp, ip->i_devvp, fsbtodb(fs, ino_to_fsba(fs, ip->i_number)));

  /* unlock the log */
  lockmgr(&mp->mnt_loglock, LK_RELEASE, NULL, NULL);
}

